#include<stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_math.h>
#include "cmdline.h"
#include "chk.h"

#define BITMASK(b)    (1 << ( (b) % CHAR_BIT))
#define BITSLOT(b)    ((b) / CHAR_BIT)
#define BITSET(a,b)   ((a)[BITSLOT(b)] |= (char)BITMASK(b))
#define BITTEST(a,b)  ((a)[BITSLOT(b)] & BITMASK(b))
#define BITCLEAR(a,b) ((a)[BITSLOT(b)] &= ~BITMASK(b))
#define BITNSLOTS(nb) ((nb + CHAR_BIT - 1) / CHAR_BIT)

void bitarray2d_print(char *bitarray, size_t bitarray_x,
		      size_t bitarray_y);

size_t bitarray2d_true_count(char *, size_t, size_t);
void bitarray2d_true_get(char *, size_t, size_t, size_t, size_t *,
			 size_t *);

void bitarray2d_true_get(bitarray, bitarray_x, bitarray_y, bitarray_j, x_j,
			 y_j)
char *bitarray;
size_t bitarray_x;
size_t bitarray_y;
size_t bitarray_j;
size_t *x_j;
size_t *y_j;
{
	size_t N = 0;
	size_t i = 0;
	size_t bitarray_num = bitarray_x * bitarray_y;
	for (i = 0; i < bitarray_num; i++) {
		if (0 != BITTEST(bitarray, i)) {
			/*
			   fprintf(stderr,"%d\t",i);
			   fprintf(stderr,"%d\n",N);
			 */
			if (N == bitarray_j) {
				*x_j = i / bitarray_y;
				*y_j = i % bitarray_y;
				/*
				   fprintf(stderr,"\n");
				 */
				return;
			}
			N++;
		}
	}
	return;
}

size_t bitarray2d_true_count(bitarray, bitarray_x, bitarray_y)
char *bitarray;
size_t bitarray_x;
size_t bitarray_y;
{
	size_t N = 0;
	size_t i = 0;
	size_t bitarray_num = bitarray_x * bitarray_y;
	for (i = 0; i < bitarray_num; i++) {
		if (0 != BITTEST(bitarray, i)) {
			/*
			   fprintf(stderr,"%d\t",i);
			   fprintf(stderr,"%d\n",N);
			 */
			N++;
		}
	}
	/*
	   fprintf(stderr,"\n");
	 */
	return N;
}

void bitarray2d_print(bitarray, bitarray_x, bitarray_y)
char *bitarray;
size_t bitarray_x;
size_t bitarray_y;
{
	size_t i;
	size_t j;
	for (i = 0; i < bitarray_x; i++) {
		for (j = 0; j < bitarray_y; j++) {
			if (0 != (BITTEST(bitarray, i * bitarray_y + j))) {
				printf("%d", 1);
			} else {
				printf("%d", 0);
			}
		}
		printf("\n");
	}
}

typedef struct {
	size_t dla_x;		/* dla x-size */
	size_t dla_y;		/* dla y-size */
	size_t step;		/* the number of randum groth */
	/* int *bitarray; */
	char *bitarray;
} param;

int main(int, char **);
void new_param(param *);
void print_param(param *);
void free_param(param *);

void new_param(p)
param *p;
{
	size_t bitarray_bit;
	size_t bitarray_byte;
	/*
	 * p->dla_x = 1000;
	 p->dla_y = 1000;
	 p->step = 1000;
	 */
	bitarray_bit = p->dla_x * p->dla_y;
	bitarray_byte = (bitarray_bit / CHAR_BIT) + 1;
	p->bitarray = (char *) chk_malloc(bitarray_byte);
	memset(p->bitarray, 0, bitarray_byte);
	return;
}

void print_param(p)
param *p;
{
	size_t dla_x = p->dla_x;
	size_t dla_y = p->dla_y;
	printf("# dla_x\t=\t%d\n", dla_x);
	printf("# dla_y\t=\t%d\n", dla_y);
	printf("# step\t=\t%d\n", p->step);
	printf("# bitarray\t=\t\n");
	/*
	   bitarray2d_print(p->bitarray, dla_x, dla_y);
	 */
	return;
}

void free_param(p)
param *p;
{
	free(p->bitarray);
	free(p);
	return;
}

int is_around_true(char *, size_t, size_t, size_t, size_t);

int is_around_true(bitarray2d, nx, ny, x, y)
char *bitarray2d;
size_t nx;
size_t ny;
size_t x;
size_t y;
{
	size_t k = x * ny + y;
	/*
	   fprintf(stderr,"%d\t",k);
	 */
	if (x == 0 && y == 0) {
		if (0 != BITTEST(bitarray2d, k + ny)) {
			/*
			   fprintf(stderr,"%d\n",k+ny);
			 */
			return 1;
		} else if (0 != BITTEST(bitarray2d, k + 1)) {
			/*
			   fprintf(stderr,"%d\n",k+1);
			 */
			return 1;
		} else {
			return 0;
		}
	} else if (x == 0 && y == ny - 1) {
		if (0 != BITTEST(bitarray2d, k + ny)) {
			/*
			   fprintf(stderr,"%d\n",k+ny);
			 */
			return 1;
		} else if (0 != BITTEST(bitarray2d, k - 1)) {
			/*
			   fprintf(stderr,"%d\n",k- 1);
			 */
			return 1;
		} else {
			return 0;
		}
	} else if (x == nx - 1 && y == 0) {
		if (0 != BITTEST(bitarray2d, k - ny)) {
			/*
			   fprintf(stderr,"%d\n",k-ny);
			 */
			return 1;
		} else if (0 != BITTEST(bitarray2d, k + 1)) {
			/*
			   fprintf(stderr,"%d\n",k+ 1);
			 */
			return 1;
		} else {
			return 0;
		}
	} else if (x == nx - 1 && y == ny - 1) {
		if (0 != BITTEST(bitarray2d, k - ny)) {
			/*
			   fprintf(stderr,"%d\n",k-ny);
			 */
			return 1;
		} else if (0 != BITTEST(bitarray2d, k - 1)) {
			/*
			   fprintf(stderr,"%d\n",k- 1);
			 */
			return 1;
		} else {
			return 0;
		}
	} else if (x == 0) {
		if (0 != BITTEST(bitarray2d, k + ny)) {
			/*
			   fprintf(stderr,"%d\n",k+ny);
			 */
			return 1;
		} else if (0 != BITTEST(bitarray2d, k + 1)) {
			/*
			   fprintf(stderr,"%d\n",k+ 1);
			 */
			return 1;
		} else if (0 != BITTEST(bitarray2d, k - 1)) {
			/*
			   fprintf(stderr,"%d\n",k- 1);
			 */
			return 1;
		} else {
			return 0;
		}
	} else if (x == nx - 1) {
		if (0 != BITTEST(bitarray2d, k - ny)) {
			/*
			   fprintf(stderr,"%d\n",k-ny);
			 */
			return 1;
		} else if (0 != BITTEST(bitarray2d, k + 1)) {
			/*
			   fprintf(stderr,"%d\n",k+ 1);
			 */
			return 1;
		} else if (0 != BITTEST(bitarray2d, k - 1)) {
			/*
			   fprintf(stderr,"%d\n",k- 1);
			 */
			return 1;
		} else {
			return 0;
		}
	} else if (y == 0) {
		if (0 != BITTEST(bitarray2d, k + ny)) {
			/*
			   fprintf(stderr,"%d\n",k+ny);
			 */
			return 1;
		} else if (0 != BITTEST(bitarray2d, k - ny)) {
			/*
			   fprintf(stderr,"%d\n",k-ny);
			 */
			return 1;
		} else if (0 != BITTEST(bitarray2d, k + 1)) {
			/*
			   fprintf(stderr,"%d\n",k+ 1);
			 */
			return 1;
		} else {
			return 0;
		}
	} else if (y == ny - 1) {
		if (0 != BITTEST(bitarray2d, k + ny)) {
			/*
			   fprintf(stderr,"%d\n",k+ny);
			 */
			return 1;
		} else if (0 != BITTEST(bitarray2d, k - ny)) {
			/*
			   fprintf(stderr,"%d\n",k-ny);
			 */
			return 1;
		} else if (0 != BITTEST(bitarray2d, k - 1)) {
			/*
			   fprintf(stderr,"%d\n",k- 1);
			 */
			return 1;
		} else {
			/*
			   fprintf(stderr,"%d\n",y);
			   fprintf(stderr,"%d\n",ny);
			 */
			return 0;
		}
	} else {
		if (0 != BITTEST(bitarray2d, k + ny)) {
			/*
			   fprintf(stderr,"%d\n",k+ny);
			 */
			return 1;
		} else if (0 != BITTEST(bitarray2d, k - ny)) {
			/*
			   fprintf(stderr,"%d\n",k-ny);
			 */
			return 1;
		} else if (0 != BITTEST(bitarray2d, k + 1)) {
			/*
			   fprintf(stderr,"%d\n",k+ 1);
			 */
			return 1;
		} else if (0 != BITTEST(bitarray2d, k - 1)) {
			/*
			   fprintf(stderr,"%d\n",k -1);
			 */
			return 1;
		} else {
			return 0;
		}
	}
	/*
	   fprintf(stderr,"# return 0\n");
	 */
	return 0;
}

void l(double, double *, double *);

void l(t, x, y)
double t;
double *x;
double *y;
{
	t -= (double) floor(t);
	if (0.00 <= t && t < 0.25) {
		*x = 2.0 / (0.25 - 0.00) * (t - 0.00) - 1.0;
		*y = -1.0;
	} else if (0.25 <= t && t < 0.50) {
		*x = 1.0;
		*y = 2.0 / (0.50 - 0.25) * (t - 0.25) - 1.0;
	} else if (0.50 <= t && t < 0.75) {
		*x = 2.0 / (0.75 - 0.50) * (t - 0.50) - 1.0;
		*y = 1.0;
	} else if (0.75 <= t && t < 1.00) {
		*x = -1.0;
		*x = 2.0 / (1.00 - 0.75) * (t - 0.75) - 1.0;
	}
	return;
}

void growth_dla3(param *);

void growth_dla3(p)
param *p;
{
	size_t step = p->step;
	const gsl_rng_type *T;
	gsl_rng *r;
	size_t dla_x = p->dla_x;
	size_t dla_y = p->dla_y;
	double u = 0.0;
	double v = 0.0;
	size_t i = 0;
	size_t x0 = dla_x / 2;
	size_t y0 = dla_y / 2;
	size_t x = x0;
	size_t y = y0;
	size_t x_min = x;
	size_t x_max = x;
	size_t y_min = y;
	size_t y_max = y;
	size_t k = x * dla_y + y;
	/*
	   printf("%d\t", k);
	   printf("%d\n", i);
	 */
	BITSET(p->bitarray, k);
	printf("# ");
	printf("x\t");
	printf("y\t");
	printf("x_min\t");
	printf("x_max\t");
	printf("y_min\t");
	printf("y_max\t");
	printf("i\n");
	printf("%d\t", x);
	printf("%d\t", y);
	printf("%d\t", x_min);
	printf("%d\t", x_max);
	printf("%d\t", y_min);
	printf("%d\t", y_max);
	printf("%d\n", i);
	gsl_rng_env_setup();
	T = gsl_rng_default;
	r = gsl_rng_alloc(T);
	for (i = 1; i < step; i++) {
		/*
		 * define the side and position(x,y) for an element incidenting into dlaclusrer cell
		 * */
		u = gsl_rng_uniform(r);
		v = gsl_rng_uniform(r);
		if (0.00 <= u && u < 0.25) {
			/* from left side */
			x = x_min - 1;
			y = (size_t) (v * (double) dla_y);
		} else if (0.25 <= u && u < 0.50) {
			/* from right side */
			x = x_max + 1;
			y = (size_t) (v * (double) dla_y);
		} else if (0.50 <= u && u < 0.75) {
			/* from bottom side */
			x = (size_t) (v * (double) dla_x);
			y = y_min - 1;
		} else if (0.75 <= u && u < 1.00) {
			/* from top side */
			x = (size_t) (v * (double) dla_x);
			y = y_max + 1;
		}
		x0 = x;
		y0 = y;
		/*
		   fprintf(stderr, "%d\t", x);
		   fprintf(stderr, "%d\t", y);
		   fprintf(stderr, "%d\n", x * dla_y + y);
		 */
		while (0 ==
		       is_around_true(p->bitarray, dla_x, dla_y, x, y)) {
			do {
				u = gsl_rng_uniform(r);
				if (0.00 <= u && u < 0.25) {
					x = x0 + 1;
					y = y0 + 0;
				} else if (0.25 <= u && u < 0.50) {
					x = x0 + 0;
					y = y0 + 1;
				} else if (0.50 <= u && u < 0.75) {
					x = x0 - 1;
					y = y0 + 0;
				} else if (0.75 <= u && u < 1.00) {
					x = x0 + 0;
					y = y0 - 1;
				}
			} while (dla_x - 1 < x || dla_y - 1 < y);
			x0 = x;
			y0 = y;
			/*
			   fprintf(stderr, "%d\t", x);
			   fprintf(stderr, "%d\t", y);
			   fprintf(stderr, "%d\n", x * dla_y + y);
			 */
		}
		/*
		   fprintf(stderr, "#end\n");
		   fprintf(stderr, "\n");
		 */
		if (x < x_min) {
			x_min = x;
		} else if (x_max < x) {
			x_max = x;
		}
		if (y < y_min) {
			y_min = y;
		} else if (y_max < y) {
			y_max = y;
		}
		k = x * dla_y + y;
		BITSET(p->bitarray, k);
		x0 = x;
		y0 = y;
		printf("%d\t", x);
		printf("%d\t", y);
		printf("%d\t", x_min);
		printf("%d\t", x_max);
		printf("%d\t", y_min);
		printf("%d\t", y_max);
		printf("%d\t", i);
		printf("%d\n", k);
	}
	gsl_rng_free(r);
	return;
}

void growth_dla2(param *);

void growth_dla2(p)
param *p;
{
	size_t step = p->step;
	const gsl_rng_type *T;
	gsl_rng *r;
	size_t dla_x = p->dla_x;
	size_t dla_y = p->dla_y;
	double u = 0.0;
	double v = 0.0;
	size_t i = 0;
	size_t x0 = dla_x / 2;
	size_t y0 = dla_y / 2;
	size_t x = x0;
	size_t y = y0;
	size_t k = x * dla_y + y;
	BITSET(p->bitarray, k);
	printf("# x\ty\ti\ti\n");
	printf("%d\t", x);
	printf("%d\n", y);
	/*
	   printf("%d\t", k);
	   printf("%d\n", i);
	 */
	gsl_rng_env_setup();
	T = gsl_rng_default;
	r = gsl_rng_alloc(T);
	for (i = 1; i < step; i++) {
		u = gsl_rng_uniform(r);
		v = gsl_rng_uniform(r);
		if (0.00 <= u && u < 0.25) {
			x = 0;
			y = (size_t) (v * (double) dla_y);
		} else if (0.25 <= u && u < 0.50) {
			x = dla_x - 1;
			y = (size_t) (v * (double) dla_y);
		} else if (0.50 <= u && u < 0.75) {
			x = (size_t) (v * (double) dla_x);
			y = 0;
		} else if (0.75 <= u && u < 1.00) {
			x = (size_t) (v * (double) dla_x);
			y = dla_y - 1;
		}
		x0 = x;
		y0 = y;
		/*
		   fprintf(stderr, "%d\t", x);
		   fprintf(stderr, "%d\t", y);
		   fprintf(stderr, "%d\n", x * dla_y + y);
		 */
		while (0 ==
		       is_around_true(p->bitarray, dla_x, dla_y, x, y)) {
			do {
				u = gsl_rng_uniform(r);
				if (0.00 <= u && u < 0.25) {
					x = x0 + 1;
					y = y0 + 0;
				} else if (0.25 <= u && u < 0.50) {
					x = x0 + 0;
					y = y0 + 1;
				} else if (0.50 <= u && u < 0.75) {
					x = x0 - 1;
					y = y0 + 0;
				} else if (0.75 <= u && u < 1.00) {
					x = x0 + 0;
					y = y0 - 1;
				}
			} while (dla_x - 1 < x || dla_y - 1 < y);
			x0 = x;
			y0 = y;
			/*
			   fprintf(stderr, "%d\t", x);
			   fprintf(stderr, "%d\t", y);
			   fprintf(stderr, "%d\n", x * dla_y + y);
			 */
		}
		/*
		   fprintf(stderr, "#end\n");
		   fprintf(stderr, "\n");
		 */
		k = x * dla_y + y;
		BITSET(p->bitarray, k);
		x0 = x;
		y0 = y;
		printf("%d\t", x);
		printf("%d\n", y);
		/*
		   printf("%d\t", k);
		   printf("%d\n", i);
		 */
	}
	gsl_rng_free(r);
	return;
}

void growth_dla(param *, size_t);

void growth_dla(p, step)
param *p;
size_t step;
{
	const size_t dla_x = p->dla_x;
	const size_t dla_y = p->dla_y;
	char *bitarray = p->bitarray;
	double u = 0.0;
	size_t N = 0;
	size_t x0 = dla_x / 2;
	size_t y0 = dla_y / 2;
	size_t x = x0;
	size_t y = y0;
	size_t i = 0;
	const gsl_rng_type *T;
	gsl_rng *r;
	gsl_rng_env_setup();
	T = gsl_rng_default;
	r = gsl_rng_alloc(T);
	i = x * dla_y + y;
	BITSET(bitarray, i);
	N = bitarray2d_true_count(bitarray, dla_x, dla_y);
	printf("# x\ty\ti\tN\n");
	printf("%d\t", x);
	printf("%d\t", y);
	printf("%d\t", i);
	printf("%d\n", N);
	while (N < step) {
		do {
			u = gsl_rng_uniform(r);
			if (0.00 <= u && u < 0.25) {
				x = x0 + 1;
				y = y0 + 0;
			} else if (0.25 <= u && u < 0.50) {
				x = x0 + 0;
				y = y0 + 1;
			} else if (0.50 <= u && u < 0.75) {
				x = x0 - 1;
				y = y0 + 0;
			} else if (0.75 <= u && u < 1.00) {
				x = x0 + 0;
				y = y0 - 1;
			}
		} while (dla_x - 1 < x || dla_y - 1 < y);
		i = x * dla_y + y;
		BITSET(bitarray, i);
		N = bitarray2d_true_count(bitarray, dla_x, dla_y);
		x0 = x;
		y0 = y;
		printf("%d\t", x);
		printf("%d\t", y);
		printf("%d\t", i);
		printf("%d\n", N);
	}
	gsl_rng_free(r);
	return;
}

static struct gengetopt_args_info args_info;


int cmd(int argc, char **argv, param * p);

int cmd(argc, argv, p)
int argc;
char **argv;
param *p;
{
	int result = 0;

	if (cmdline_parser(argc, argv, &args_info) != 0) {
		result = 1;
		goto stop;
	}
	if (args_info.conf_in_given) {
		if (cmdline_parser_configfile
		    (args_info.conf_in_arg, &args_info, 0, 0, 1) != 0) {
			result = 1;
			goto stop;
		}
	}

	if (cmdline_parser_file_save
	    (args_info.conf_out_arg, &args_info) == EXIT_FAILURE)
		result = 1;
	else
		printf("# saved configuration file %s\n",
		       args_info.conf_out_arg);
      stop:

	p->dla_x = (size_t) args_info.nx_arg;
	p->dla_y = (size_t) args_info.ny_arg;
	p->step = (size_t) args_info.step_arg;
	if (p->dla_x * p->dla_y < p->step) {
		p->step = p->dla_x * p->dla_y;
	}
	cmdline_parser_free(&args_info);

	return result;
}

int main(argc, argv)
int argc;
char **argv;
{
	param *p;
	p = (param *) chk_calloc(1, sizeof(param));
	cmd(argc, argv, p);
	new_param(p);
	print_param(p);
	/*
	 * growth_dla(p, 100000);
	 growth_dla2(p);
	 * */
	growth_dla3(p);
	/*print_param(p); */
	free_param(p);
	return 0;
}
